﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Collections;

namespace jordan_svd
{
    /**
     * Class that solves the SVD
     */ 
    public class qrbsvd
    {

        public double [,] U;
        public double [,] S;
        public double[,] V;
        public double[] SA;

        // The dimensions of the input array
        protected int[] mdim;
        protected int count;

        // we define the tolerance
        protected double tol = 1.0E-13; 

        protected int maxiter;

        protected double [,] qrQ;
        protected double[,] qrS;

        // The row dimension of the input matrix
        public int rowDimension;
        //The column dimension of the input matrix
        public int columnDimention;

        /**
         * The default constructor
         */ 
        public qrbsvd()
        {
            mdim = new int[2];
        }

        /**
         * Returns the max dimension.
         */
        protected int maxMdim()
        {
            if (mdim[0] > mdim[1])
            {
                return mdim[0];
            }
            return mdim[1];
        }

        /**
         * Creates and returns an identity matrix mXn
         */
        protected double[,] identity(int m, int n)
        {
            double[,] result = new double[m, n];
			for (int i = 0; i < m; i++)
			{
				for (int j = 0; j < n; j++)
				{
					result[i,j] = (i == j ? 1.0 : 0.0);
				}
			}
			return result;
        }

        /**
         * Returns the transpose of the specified matrix
         */ 
        public double[,] transpose(double[,] inm)
        {
            int n = inm.GetLength(1);
            int m = inm.GetLength(0);
            double[,] result = new double[n,m];

            for (int i = 0; i < m; i++)
            {
                for (int j = 0; j < n; j++)
                {
                    result[j,i] = inm[i,j];
                }
            }
            return result;

        }

        /**
         * Performs the QR of the specified matrix.
         */ 
        public void qr(double[,] toqr)
        {
            int m = toqr.GetLength(0);
            int n = toqr.GetLength(1);
            double [] tau = new double[Math.Min(m,n)];

            alglib.ortfac.rmatrixqr(ref toqr, m, n, ref tau);
            m = toqr.GetLength(0);
            n = toqr.GetLength(1);
            int minq = Math.Min(m,n);
            qrQ = new double[minq,minq];
            alglib.ortfac.rmatrixqrunpackq ( toqr , m,n,tau, m, ref qrQ );
            qrS = new double[m, n];
            alglib.ortfac.rmatrixqrunpackr (toqr,m,n,ref qrS);

        }

        /**
         * Multiplies the specified arrays.
         */ 
        public double[,] mulArray(double[,] a, double[,] b)
        {
            int m = a.GetLength(0);
            int n = b.GetLength(1);
            int k = a.GetLength(1);
            double[,] c = new double[m, n];
            alglib.rmatrixgemm(m, n, k, 1, a, 0, 0, 0, b, 0, 0, 0, 0, ref c, 0, 0);
            return c;
        }

        /**
         *  Returns the norm of the specified matrix.
         */ 
        protected double norm(double [] z)
        {
            double result = 0;
            for (int i = 0; i < z.Length; i++)
            {
                result = result + (z[i] * z[i]);
            }

            return Math.Sqrt(result);
        }

        /**
         * Sets and returns the diagonal of the specified matrix to 0 (triu)
         */
        protected double [,] triu (double [,] a) 
        {
            int m = a.GetLength(0);
            int n = a.GetLength(1);
            double[,] result = new double[m, n];

            for (int i = 0; i < m; i++)
            {
                for (int j = 0; j < n; j++)
                {
                    if (i == j)
                    {
                        result[i, j] = 0;
                    }
                    else
                    {
                        result[i, j] = a[i, j];
                    }
                }
            }
            return result;
        }

        /**
         * Returns matrix that contains the diagonal of the specified matrix.
         */
        public double[] diag(double[,] a)
        {
            int m = a.GetLength(0);
            int n = a.GetLength(1);
            int dg = Math.Min(m, n);
            double[] result = new double[dg];
            for (int i = 0; i < dg; i++)
            {
                result[i] = a[i, i];
            }
            return result;
        }

        /**
         * Returns the elements of the specified matrix as vector.
         */ 
        protected double[] eano(double[,] a)
        {
            int m = a.GetLength(0);
            int n = a.GetLength(1);
            int total = m * n;
            double[] result = new double[total];
            int k =0;
            for (int j = 0; j < n; j++)
            {
                for ( int i= 0 ; i < m ; i++) {
                    result[k] = a[i, j];
                    k++;
                }
            }

            return result;
        }

        /**
         * Creates and returns a zero matrix mXn
         */ 
        protected double[,] zeros(int m, int n)
        {
            double[,] result = new double[m,n];
            for (int i = 0; i < m; i++)
            {
                for (int j = 0; j < n; j++)
                {
                    result[i, j] = 0;
                }
            }
            return result;
        }

        /**
         * Multiplies the specified column with the specified value.
         */ 
        protected void mulColumn(double[,] a, double val , int n)
        {
            int m = a.GetLength(0);
            
            for (int i = 0; i < m; i++)
            {
                a[i, n] = a[i, n] * val;
            }
        }

        /**
         * Makes canonical the S.
         */ 
        protected void createSFromAlg(double[] s)
        {
            S = new double[mdim[0], mdim[1]];
            for (int i = 0; i < mdim[0]; i++)
            {
                S[i, i] = s[i];
            }
        }

        /**
         * Solves the SVD using the Alg library.
         */ 
        public void solveAlgSvd(double[,] A)
        {
            int m = A.GetLength(0);
            int n = A.GetLength(1);

            mdim[0] = m;
            mdim[1] = n;

            double [] w;
            double [,] u;
            double [,] vt;


            alglib.rmatrixsvd(A, m, n, 2, 2, 2, out w, out u, out vt);

            this.U = u;
            this.V = this.transpose(vt);
            this.SA = w;

            createSFromAlg(SA);

        }

        /**
         * Solves the SVD.
         */ 
        public void solveSvd(double[,] A)
        {
            
            double err = 1.0E308;
            
            mdim[0] = A.GetLength(0);
            mdim[1] = A.GetLength(1);

            rowDimension = mdim[0];
            columnDimention = mdim[1];

            maxiter = 100 * maxMdim();
            count = 0;

            U = identity(mdim[0], mdim[0]);
            S = transpose(A);
            V = identity(mdim[1], mdim[1]);

            double[,] Q;

            while ((err > tol) && (count < maxiter))
            {
                qr(transpose(S));
                S = qrS;
                Q = qrQ;
                U = mulArray(U, Q);
                qr(transpose(S));
                S = qrS;
                Q = qrQ;
                V = mulArray(V, Q);

                double[,] e;
                e = triu(S);
                double E , F;
                E = norm(eano(e));
                F = norm(diag(S));

                if (F == 0)
                {
                    F = 1.0;
                }

                err = E / F;
                count = count + 1;
            } // end while tol and count


            double[] D;
            D = diag(S);
            S = zeros(mdim[0], mdim[1]);
            double dn;
            for (int n = 0; n < D.Length; n++)
            {
                dn = D[n];
                S[n, n] = Math.Abs(dn);
                if (dn < 0)
                {
                    mulColumn(U,(-1),n);
                }
            }
            
        }


        /**
         * Returns the Rank
         */ 
        public int Rank()
        {
            int r = 0;
            int m = S.GetLength(0);
            int n = S.GetLength(1);
            for (int i = 0; i < m; i++)
            {
                for (int j = 0; j < n; j++)
                {
                    if (S[i,j] > tol)
                    {
                        r++;
                    }

                }
            }
            return r;
        }
    }
}
